/******************************************************************************
Name: bridge.C

Description: This file holds the main function of the BOINC->Grid bridge and
	     the infinite loop, where all the 'work' is done. It also holds
	     a few functions to handle signals. This is an attempt to avoid
	     inconsistency in the database. 
	     
Authors: Jakob Gregor Pedersen & Christian Ulrik Sttrup.
******************************************************************************/

#include <time.h>
#include <signal.h>
#include <unistd.h>
#include "boinc_db.h"
#include "bridge_db.h"
#include "bridge_util.h"
#include "backend_lib.h"
#include "bridge_grid.h"
#include "util.h"
#include "sched_util.h"
#include "sched_msgs.h"
#include "edg/workload/logging/client/JobStatus.h" //enums for status of a Grid job.

using namespace edg::workload::logging::client;


//Defines.
#define LIMIT 100 //Maximum returned results from a database query.
#define STATUS_INTERVAL 1 //How often (after an initial period) we query the Grid for the status of jobs (in minutes).

//Global variabels
SCHED_CONFIG config; //Contains the configuration settings from config.xml.
volatile sig_atomic_t sigexit = 0; //Is set if we are to exit the program.


/******************************************************************************
Name: sighandler

Function: Sets the global variabel sigexit, which is checked at convenient
	  times in bridge_loop, and prints an exit messages.  

Called by: the system but installed for signal handling by install_sighandler.
******************************************************************************/
void sighandler(int sig){
	sigexit = 1;
	write(1, "Shutting down...\n", 18); //write is reentrant - printf is not.
	return; 
}


/******************************************************************************
Name: install_sighandler

Function: Installs sighandler to handle SIGINT, SIGHUP, SIGTERM and SIGQUIT.  

Called by: bridge_loop.
******************************************************************************/
int install_sighandler(){
	struct sigaction act;
	int retval;
	
	act.sa_handler = sighandler;
	retval = sigfillset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	retval += sigaction(SIGINT, &act, 0);
	retval += sigaction(SIGHUP, &act, 0);
	retval += sigaction(SIGTERM, &act, 0);
	retval += sigaction(SIGQUIT, &act, 0);
	
	return retval;
}


/******************************************************************************
Name: bridge_loop

Function: This is the main loop of the bridge that loops until instructed to 
	  exit. The steps of each loop is as follows:
	  - Check for possibly finished jobs.
	  - Get the status of these possibly finished jobs.
	  - Get output and update the database for finished jobs.
	  - Check for jobs that are about to violate their time to finish.
	  - Create a new result in the result table for these jobs
	  - Create needed files (wrapper-script, jdl-file etc.) for these jobs.
	  - Submit the jobs to the Grid.
	  - Create new entries in the grid_job table to record the submissions.  

Called by: main.
******************************************************************************/
void bridge_loop(){
	GRID_DB_WORK_ITEM gwi;
	GRID_APP_VERSION gav;
	GRID_USER gu;
	int retval, loop_time, previous_appid;
	time_t start_loop;
	int status, done_code, exit_code;
	R_RSA_PRIVATE_KEY key;
	char path[256];
	
	//Get the private upload key to be used when the 'xml_doc_in' for new result are generated.	
	sprintf(path, "%s/upload_private", config.key_dir);
	retval = read_key_file(path, key);
	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Can't read key: '%s'!\n", path);
		exit(EXIT_FAILURE);
	}
	
	retval = create_account_xml();
	
	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Could not find/create account-xml-file!\n");
		exit(EXIT_FAILURE);
	}
	
	//Get information about the Grid 'user' from the database - This is purely cosmetic.
	gu.lookup();
	
	retval = install_sighandler();
	
	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to install signal handler!\n");
	}
	
		
	while (1) //The infinite loop.
	{
		if(sigexit) //Check for exit-signals.
			exit(EXIT_SUCCESS);
		
		start_loop = time(0); //Record the time we started looping, used to check if we are using to much time.
		
		//Check to see if we have any jobs that might be finished.
		retval = gwi.get_possibly_finished_jobs(LIMIT); //maybe no limit?
		if (retval)
			log_messages.printf(SCHED_MSG_LOG::DEBUG, "No possibly finished jobs found.\n");
		
		//Check the status of each possibly finished job and act appropriately.
		while(!retval)
		{
			if(sigexit) //Check for exit-signals
				exit(EXIT_SUCCESS);
			
			retval = get_job_status(gwi.grid_jobid, status, done_code, exit_code); //Get the status of the job from Grid.
			if (retval){ //If get_job_status fails for the present job try the next one.
				retval = gwi.get_possibly_finished_jobs(LIMIT);
				continue;
			}
			
			switch (status){ //Act appropriately to the job status - BTW a failed job is not resubmitted.
				case JobStatus::DONE:
					if(exit_code){ //If the exit code is not 0 the job failed.
						log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Result %d failed with exit-code '%d'!\n"
								, gwi.resultid, exit_code);
						retval = get_job_output(gwi.grid_jobid, config.upload_dir); //put the output in the upload directory
						if (retval)
							log_messages.printf(SCHED_MSG_LOG::CRITICAL,
								"Failed to get output from result %d!", gwi.resultid);
						gwi.set_job_failed(exit_code);
						break;
					}
					if(!(done_code == JobStatus::DONE_CODE_OK)){ //Usually these jobs recover so they are given second chances!
						if (gwi.report_deadline < time(0)){
							log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Result %d exceeded its report-deadline, it will be consider as failed!\n"
								, gwi.resultid);
							gwi.set_job_failed();
							break;
						}
						log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Result %d failed with done-code '%d'!"
								" Giving it a second chance!\n"
								, gwi.resultid, done_code);
						gwi.set_status_time(time(0)+(STATUS_INTERVAL*60)); //Giving it (unlimited) second chances!!
						if (retval)
							log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Setting the status time for result %d failed!\n", gwi.resultid);
						break;
					}
					log_messages.printf(SCHED_MSG_LOG::DEBUG,
                        					"Result %d is done.\n", gwi.resultid);
					retval = get_job_output(gwi.grid_jobid, config.upload_dir); //put the output in the upload directory
					if (retval){
						log_messages.printf(SCHED_MSG_LOG::CRITICAL,
								"Failed to get output from result %d!", gwi.resultid);
						gwi.set_job_failed();
						break;
					}
					retval = gwi.set_job_finished();
					if (retval)
						log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Setting the status of result %d to 'finished' failed with error#: %d!\n"
								, gwi.resultid, retval);	
					break;
					
				case JobStatus::CLEARED: //We shouldn't get this status because once we have collected the output the job status 
							 //is never queried again.
					log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Result %d returned status 'cleared'. This is not supposed to happen!\n"
								, gwi.resultid);
					gwi.handle_cleared_status();
					break;
				
				case JobStatus::ABORTED: //We shouldn't get this either (unless aborted by the Grid that is).
					log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Result %d has been aborted!\n", gwi.resultid);
					gwi.set_job_failed();
					break;
					
				case JobStatus::CANCELLED: //We shouldn't get this either.
					log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Result %d has been cancelled!\n", gwi.resultid);
					gwi.set_job_failed();
					break;
					
				case JobStatus::PURGED: //We shouldn't get this either.
					log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Result %d has been purged!\n", gwi.resultid);
					gwi.set_job_failed();
					break;
					
				default: //Either UNDEF, SUBMITTED, WAITING, READY, SCHEDULED, RUNNING or UNKNOWN, ie. hopefully not finished...
					if (gwi.report_deadline < time(0)){
						log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        				"Result %d exceeded its report-deadline, it will be consider as failed!\n"
							, gwi.resultid);
						gwi.set_job_failed();
						break;
					}
					log_messages.printf(SCHED_MSG_LOG::DEBUG, 
						"Result %d is not finished yet.\n", gwi.resultid);
					gwi.set_status_time(time(0)+(STATUS_INTERVAL*60));
					if (retval)
						log_messages.printf(SCHED_MSG_LOG::CRITICAL,
                        					"Setting the status time for result %d failed!\n", gwi.resultid);
					break;
			}
			
			retval = gwi.get_possibly_finished_jobs(LIMIT); //Look at the next possibly finished job.
		}
		
		if(sigexit) //Check for exit-signals.
			exit(EXIT_SUCCESS);		
		
		
		//Now we are done with the jobs already submitted and go on to submitting new jobs that are about to violate their set time to finish.
		
		previous_appid = 0; //To force checking for new app_version each time around the main loop.
		
		//Get the jobs that need to be submitted to Grid.
		retval = gwi.get_grid_jobs(LIMIT, config.grid_fpops, time(0)); //maybe no limit?
		if (retval){
			log_messages.printf(SCHED_MSG_LOG::DEBUG, "No jobs for Grid found.\n");
			loop_time = (time(0) - start_loop);
			if (loop_time >= config.bridge_loop_interval){
				log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Time spent looping is bigger than the set 'bridge_loop_interval'"
									     " - Struggling to keep up with the workload!!!\n");
				continue;
			}
			
			sleep(config.bridge_loop_interval-loop_time); 
			continue;
		}
		
		while(!retval)
		{
			if(sigexit) //Check for exit-signals.
				exit(EXIT_SUCCESS);
			
			log_messages.printf(SCHED_MSG_LOG::DEBUG, "Submitting a result for workunit %d to Grid.\n", gwi.wu.id);
						
			if (previous_appid != gwi.wu.appid){ //Don't need to lookup everytime.
				gav.lookup(gwi.wu.appid); //If we haven't done it already lookup the latest version of the app in question.
				
				//Prepare the xml_doc for when the client_state.xml file is created.
				retval = str_insert_after_match(gav.xml_doc, "<file_info>", "\n\t<status>1</status>", sizeof(gav.xml_doc));
				if(!retval){
					log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to prepare app_version:xml_doc for workunit %d.\n", gwi.wu.id);
					previous_appid = 0; //Need another lookup!
					retval = gwi.get_grid_jobs(LIMIT, config.grid_fpops, time(0));
					continue;
				}
				previous_appid = gwi.wu.appid;
			}
			//Generate xml_doc_in for the new result which will be inserted into the result table.
			retval = generate_result_xml_doc_in(gwi, key, config.upload_url);
			
			if(retval){
				log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to generate xml_doc_in for "
									     "result connected to workunit %d.\n", gwi.wu.id);
				retval = gwi.get_grid_jobs(LIMIT, config.grid_fpops, time(0));
				continue;
			}
			//Create the files needed for the job (wrapper_script, client_state.xml etc.)			
			retval = create_files_for_job(gwi, gav, config.VO, config.grid_retrycount);
			
			if(retval){
				log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to create needed files for workunit %d.\n", gwi.wu.id);
				retval = gwi.get_grid_jobs(LIMIT, config.grid_fpops, time(0));
				continue;
			}
			
			//add_result uses 'replace' instead of 'insert', which should remove the need to roll back if the rest fails!
			retval = gwi.add_result(gu, gav.version, time(0));
			
			if (retval)
				log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to generate result for workunit %d\n", gwi.wu.id);
				//Maybe stop this from happening again?
			
			else {
				log_messages.printf(SCHED_MSG_LOG::DEBUG, "Generated result %d for submission to Grid.\n", gwi.resultid);
			
				//Submit the job to Grid.
				retval = submit_result(config, gwi.wu.name, gwi.wu.id, gwi.grid_jobid, sizeof(gwi.grid_jobid));
				if (retval){
					retval = gwi.get_grid_jobs(LIMIT, config.grid_fpops, time(0));
					continue;
					//Maybe stop this from happening again?
				}
				log_messages.printf(SCHED_MSG_LOG::DEBUG, "Submitted result %d to Grid.\n", gwi.resultid);
				
				retval = gwi.add_job_to_grid_table(config.grid_fpops); //Add an entry the grid_job table to record the submission.
				if (retval){
					log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
							"Failed to record the submission of result '%d' to the database! VERY BAD!!!\n", gwi.resultid);
					//maybe stop this from happening again?
				}
			}		
			retval = gwi.get_grid_jobs(LIMIT, config.grid_fpops, time(0));
		}

		loop_time = (time(0) - start_loop); //Be nice, don't use all the cpu-time looping.
		if (loop_time >= config.bridge_loop_interval){
			log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Time spent looping is bigger than the set 'LOOP_INTERVAL'"
								     " - Struggling to keep up with the workload!!!\n");
			continue;
		}
		
		if(sigexit) //Check for exit-signals.
			exit(EXIT_SUCCESS);
			
		sleep(config.bridge_loop_interval-loop_time); 
	}
}


/******************************************************************************
Name: main

Function: Parses the config-file, sets the debug level and open the database.
	  After this the function brigde_loop with the infinite loop is called.   

Called by: the system.
******************************************************************************/
int main(int argc, char** argv) {
 	int i, retval;
          
	retval = config.parse_file("..");
            
  	if (retval) {
        	log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Can't parse config file\n");
        	exit(EXIT_FAILURE);
    	}

    	for (i=1; i<argc; i++) { //Only the -d switch is recognized.
        	if (!strcmp(argv[i], "-d")) {
            	log_messages.set_debug_level(atoi(argv[++i]));
        	}
    	}
    
    	log_messages.printf(SCHED_MSG_LOG::NORMAL, "Starting...\n");

    	retval = boinc_db.open(config.db_name, config.db_host, config.db_user, config.db_passwd);
    	if (retval) {
        	log_messages.printf(SCHED_MSG_LOG::CRITICAL, "boinc_db.open: %d; %s\n", retval, boinc_db.error_string());
        	exit(EXIT_FAILURE);
    	}
    	bridge_loop();
    	
	return 0;
}
