/******************************************************************************
Name: bridge_db.C

Description: This file holds all the functions that access the database, 
	     whether just to lookup information or to change the contents
	     of the database. It contains all of the functions in the
	     'GRID_DB_WORK_ITEM', 'GRID_APP_VERSION' and 'GRID_USER' classes.
	     
Authors: Jakob Gregor Pedersen & Christian Ulrik Sttrup.
******************************************************************************/

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <cstdlib>
#include <cstring>
#include <time.h>
#include "util.h"
#include "mysql.h"
#include "boinc_db.h"
#include "result_state.h"
#include "bridge_db.h"
#include "bridge_util.h"
#include "backend_lib.h"
#include "sched_msgs.h"
#include "result_state.h"
#include "parse.h"
#include "md5_file.h"


/******************************************************************************
Name: add_job_to_grid_table

Function: Inserts a newly submitted job into the 'grid_job' table.  

Called by: bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::add_job_to_grid_table(double grid_fpops_est){
	char query[MAX_QUERY_LEN];
	int retval;
	int report_deadline;
	time_t now;
	
	now = time(0);
	
	//Set the status_time (when we expect the job to finish) to now plus expected execution-time.
	status_time = now+(int)(wu.rsc_fpops_est/grid_fpops_est);
	//Set report_deadline to now plus ten times the expected execution time.
	report_deadline = now+(10*(int)(wu.rsc_fpops_est/grid_fpops_est));
	
	sprintf(query,
    		"INSERT INTO grid_job (grid_jobid, resultid, workunitid, status_time, report_deadline) " 
    		"VALUES ('%s', %d, %d, %d, %d)",
    		grid_jobid, resultid, wu.id, status_time, report_deadline);
	
    	retval = boinc_db.do_query(query);
    	if (retval) return mysql_errno(boinc_db.mysql);
	
	return 0;
}


/******************************************************************************
Name: add_result

Function: Inserts a new result into the 'result' table. If result with identical
	  unique fields exist already, it will overwritten (updated).
	  It also stores the id of the new result in the 'resultid' variabel.  

Called by: bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::add_result(GRID_USER gu, int app_version, time_t now){
    	char query[MAX_QUERY_LEN];
    	int retval;
    	MYSQL_ROW row;
    	CURSOR localcursor; //Uses a local cursor because the global one might be in use already.
            
	//Uses 'replace' to overwrite a previous entry if such one is present.
    	sprintf(query,
    		"REPLACE INTO result (create_time, workunitid, server_state, outcome, client_state, hostid, userid, " 
		"report_deadline, sent_time, received_time, name, cpu_time, xml_doc_in, batch, "
    		"file_delete_state, validate_state, claimed_credit, granted_credit, opaque, random, app_version_num, appid, exit_status, teamid) "
    		"VALUES (%d, %d, %d, %d, %d, %d, %d, %d, %d, 0, '%s_grid', 0, '%s', 0, %d, "
    		"%d, 0, 0, 0, 7777777, %d, %d, 0, %d)",
   		(int)now, wu.id, RESULT_SERVER_STATE_IN_PROGRESS, RESULT_OUTCOME_INIT, RESULT_NEW, gu.hostid, gu.userid, MAXIMUM_DB_INT, 
		(int)now, wu.name, xml_doc_in, FILE_DELETE_INIT, VALIDATE_STATE_INIT, app_version, wu.appid, gu.teamid);

    	retval = db->do_query(query);
    	if (retval) return mysql_errno(db->mysql);
    
    	safe_strcpy(query,"SELECT LAST_INSERT_ID()"); //Getting the id of the newly inserted result.
    
    	retval = db->do_query(query);
    	if (retval) return mysql_errno(db->mysql);
    
    	localcursor.rp = mysql_store_result(db->mysql);
    	if (!localcursor.rp) return mysql_errno(db->mysql);
    
    	row = mysql_fetch_row(localcursor.rp);
    	if (!row) {
        	mysql_free_result(localcursor.rp);
		return 1;
    	}
    
    	resultid=atoi(row[0]); //Storing the id.
    	return 0;
}


/******************************************************************************
Name: delete_files

Function: Deletes all the files from a job (except for the output of course).  

Called by: set_job_finished.
******************************************************************************/
void GRID_DB_WORK_ITEM::delete_files(){
	char filename[256];
				
	//Delete the file containing the stderr output.
	sprintf(filename, "../upload/%s_std.err", wu.name);
	unlink(filename);
		
	//Delete the file containing the stdout output.
	sprintf(filename, "../upload/%s_std.out", wu.name);
	unlink(filename);
		
	//Delete the jdl-file.
	sprintf(filename, "../grid/jobs/%s.jdl", wu.name);
	unlink(filename);
		
	//Delete the client_state file.
	sprintf(filename, "../grid/jobs/%s_client_state.xml", wu.name);
	unlink(filename);
		
	//Delete the wrapper script.
	sprintf(filename, "../grid/jobs/%s_wrapper_script", wu.name);
	unlink(filename);
	
	return;
}


/******************************************************************************
Name: get_grid_jobs

Function: Gets one workunit from the database. This workunit has less than 2
	  times the expected execution time left, before it break its 
	  finish_before time, the finish_before time is not zero 
	  (meaning QoS has been enabled) and it has not already been submitted
	  to Grid. Besides this there is a linux version of the app required
	  to process the workunit. The limit parameter only set the number of
	  returned jobs per query not the total number of returned jobs, which
	  is unlimited.

Called by: brigde_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::get_grid_jobs(int limit, double grid_fpops_est, time_t now){
    	char query[MAX_QUERY_LEN];
    	int retval;
    	MYSQL_ROW row;

    	if (!cursor.active) {
   	    	sprintf(query,
            		"SELECT HIGH_PRIORITY DISTINCT workunit.* FROM workunit LEFT JOIN grid_job ON workunit.id = grid_job.workunitid, app_version, platform "
            		"WHERE workunit.appid = app_version.appid AND app_version.platformid = platform.id " 
	    		"AND workunit.assimilate_state = %d AND platform.name LIKE '%%linux%%' AND grid_job.workunitid IS NULL "
	    		"AND (2*(workunit.rsc_fpops_est/%.0f))>= (workunit.finish_before - %d) " 
	    		"AND workunit.finish_before != 0 " //We don't want non-QoS enabled workunits.
            		"ORDER BY workunit.finish_before ASC LIMIT %d", //'ORDER BY workunit.finish_before' to get the workunits having to finish soonest first. 
            		ASSIMILATE_INIT, grid_fpops_est, (int)now, limit);
	
		retval = db->do_query(query);
        	if (retval) return mysql_errno(db->mysql);
        
		cursor.rp = mysql_store_result(db->mysql);
        	if (!cursor.rp) return mysql_errno(db->mysql);
		
		returned_grid_jobs = mysql_num_rows(cursor.rp); //Note the number of returned jobs.
		
        	cursor.active = true;
    	}
      
    	row = mysql_fetch_row(cursor.rp);
    	if (!row) {
        	mysql_free_result(cursor.rp);
        	cursor.active = false;
		if (returned_grid_jobs == limit) //If the number of returned jobs from the last query is equal to the limit
						 //there  might be more jobs waiting, so query again.
			return get_grid_jobs(limit, grid_fpops_est, now);
        	return 1;
    	} 
	
	parse(row); //Store the workunit's properties in the GRID_DB_WORK_ITEM's variables.
        return 0;
}


/******************************************************************************
Name: get_possibly_finished_jobs

Function: Gets one job from the 'grid_job' table who's status_time is up.
	  It also gets the 'xml_doc_out' field from the associated entry in 
	  the 'result' table and the 'name' from the 'workunit' tabel. The 
	  first 'extra' field gets used when the 'xml_doc_out', which is needed
	  to record the finishing of the result, is generated and the second
	  'extra' field is used to get the filenames of the output files and
	  the grid specific job files for garbage collection. The limit 
	  parameter only set the number of returned jobs per query not the 
	  total number of returned jobs, which is unlimited.

Called by: bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::get_possibly_finished_jobs(int limit){
    	char query[MAX_QUERY_LEN];
   	 int retval;
    	MYSQL_ROW row;

    	if (!cursor.active) { //We can safely use the global cursor because get_grid_jobs finishes with it before, this function is run.
   	   	sprintf(query,
            		"SELECT HIGH_PRIORITY grid_job.*, result.xml_doc_in, workunit.name FROM grid_job, result, workunit "
            		"WHERE grid_job.status_time <= %d AND result.id = grid_job.resultid AND workunit.id = grid_job.workunitid " 
	    		"LIMIT %d",
           		(int) time(0), limit);
	
		retval = db->do_query(query);
        	if (retval) return mysql_errno(db->mysql);
        
		cursor.rp = mysql_store_result(db->mysql);
        	if (!cursor.rp) return mysql_errno(db->mysql);
        
		returned_possibly_finished_jobs = mysql_num_rows(cursor.rp); //Note the number of returned jobs.
	
		cursor.active = true;
    	}
      
    	row = mysql_fetch_row(cursor.rp);
    	if (!row) {
       		mysql_free_result(cursor.rp);
        	cursor.active = false;
		if (returned_possibly_finished_jobs == limit) //If the number of returned jobs from the last query is equal to the limit
						 	      //there  might be more jobs waiting, so query again.
			return get_possibly_finished_jobs(limit);
        	return 1;
    	}
	
	parse_grid(row); //Store the grid_job's properties in the GRID_DB_WORK_ITEM's variables.
    	return 0;
}


/******************************************************************************
Name: handle_cleared_status

Function: Checks if the result's server_state is 'over' and if not sets the 
	  job-status to finished. If the server_state is 'over' then the
	  status_time is set to max, since we don't need to look at this job
	  again.   

Called by: bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::handle_cleared_status(){
	char query[MAX_QUERY_LEN];
	int retval;
	MYSQL_ROW row;
	CURSOR localcursor; //Uses a local cursor since class cursor could be in use when this function is called.
	
	sprintf(query, "SELECT id FROM result WHERE id = %d AND server_state = %d",
		resultid, RESULT_SERVER_STATE_OVER);
	
	retval = db->do_query(query);
	if (retval) return mysql_errno(db->mysql);
        
	localcursor.rp = mysql_store_result(db->mysql);
        if (!localcursor.rp) return mysql_errno(db->mysql);
	
	row = mysql_fetch_row(localcursor.rp);
	if (!row){ //The server_state was not 'over'.
		mysql_free_result(localcursor.rp);
		return set_job_finished();
	}
	mysql_free_result(localcursor.rp);
	return set_status_time(MAXIMUM_DB_INT);
}


/******************************************************************************
Name: parse

Function: Parses a row from the 'workunit' table and stores it in the 
	  variables of the object.  

Called by: get_grid_jobs.
******************************************************************************/
void GRID_DB_WORK_ITEM::parse(MYSQL_ROW& r) {
    	int i=0;
    	memset(this, 0, sizeof(WORK_ITEM));
    	wu.id=atol(r[i++]);
    	wu.create_time = safe_atoi(r[i++]);
    	wu.appid = safe_atoi(r[i++]);
    	strcpy2(wu.name, r[i++]);
    	strcpy2(wu.xml_doc, r[i++]);
    	wu.batch = safe_atoi(r[i++]);
    	wu.rsc_fpops_est = atof(r[i++]);
    	wu.rsc_fpops_bound = atof(r[i++]);
    	wu.rsc_memory_bound = atof(r[i++]);
    	wu.rsc_disk_bound = atof(r[i++]);
    	wu.need_validate = safe_atoi(r[i++]);
    	wu.canonical_resultid = safe_atoi(r[i++]);
    	wu.canonical_credit = atof(r[i++]);
    	wu.transition_time = safe_atoi(r[i++]);
    	wu.delay_bound = safe_atoi(r[i++]);
    	wu.error_mask = safe_atoi(r[i++]);
    	wu.file_delete_state = safe_atoi(r[i++]);
    	wu.assimilate_state = safe_atoi(r[i++]);
    	wu.workseq_next = safe_atoi(r[i++]);
    	wu.opaque = atof(r[i++]);
   	wu.min_quorum = safe_atoi(r[i++]);
    	wu.target_nresults = safe_atoi(r[i++]);
    	wu.max_error_results = safe_atoi(r[i++]);
    	wu.max_total_results = safe_atoi(r[i++]);
    	wu.max_success_results = safe_atoi(r[i++]);
    	strcpy2(wu.result_template_file, r[i++]);
    	wu.finish_before = safe_atoi(r[i]);
}


/******************************************************************************
Name: parse_grid

Function: Parses a row from the 'grid_jobs' table plus the xml_doc_out field
	  from the associated entry in the 'result' table and the workunit name
	  and stores them in the variables of the object.  

Called by: get_possibly_finished_jobs.
******************************************************************************/
void GRID_DB_WORK_ITEM::parse_grid(MYSQL_ROW& r) {
    	strcpy2(grid_jobid, r[0]);
    	resultid = safe_atoi(r[1]);
    	wu.id = atol(r[2]);
    	status_time = atoi(r[3]);
	report_deadline = atoi(r[4]);
    	strcpy2(xml_doc_in, r[5]);
	strcpy2(wu.name, r[6]);
}


/******************************************************************************
Name: set_job_failed

Function: Updates the 'result' table to reflect the failure of a job.
	  Tries to store the contents of the stdout and stderr in the 
	  'stderr_out' field of the result for further information on the 
	  failure. After updating the 'result' table it sets the 'status_time'
	  of job to 'MAXIMUM_DB_INT' to insure that the job is not picked up
	  by get_possibly_finished_jobs.  

Called by: set_job_finished and bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::set_job_failed(int exit_status){
	char query[MAX_QUERY_LEN];
	char stderr_out[LARGE_BLOB_SIZE];
	char stderr[LARGE_BLOB_SIZE];
	char stdout[LARGE_BLOB_SIZE];
	char filename[256];
	int retval, chars;
	FILE *file;
	
	stderr[0] = 0; //initialise stderr.
	stdout[0] = 0; //initialise stdout.
	
	sprintf(filename, "../upload/%s_std.err", wu.name); //Getting the name of the file containing the stderr output.
	
	file = fopen(filename,"r");
	if(!file){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL,
			"Failed to open '%s' to read error information!\n", filename);
	}
	else{
		chars = fread(stderr, 1, (LARGE_BLOB_SIZE - 155), file); //155 is the length of stderr_out without stdout and stderr!
		fclose(file);
	}
		
	sprintf(filename, "../upload/%s_std.out", wu.name); //Getting the name of the file containing the stdout output.
	
	file = fopen(filename,"r");
	if(!file){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL,
			"Failed to open '%s' to read std out information!\n", filename);
	}
	else{
		fread(stdout, 1, (LARGE_BLOB_SIZE - 155 - chars), file);
		fclose(file);
	}

	delete_files(); //Garbage collect.
	
	//Generating the 'stderr_out' for the result.
	sprintf(stderr_out,
		"<core_client_version>Grid version</core_client_version>\n"
		"<message>\nstderr:\n%s\nstdout:\n%s\n"
		"</message>\n"
		"<active_task_state>0</active_task_state>\n"
		"<signal>0>/signal>\n"
		, stderr, stdout);
		
	str_replace(stderr_out, '\'', ''); //replace ' with  to avoid the following query failing because stderr_out is quoted with '.	
	
	sprintf(query,
		"UPDATE result SET server_state = %d, outcome = %d, client_state = %d, "
		"stderr_out = '%s', file_delete_state = %d, "
		"validate_state = %d, exit_status = %d WHERE id = %d"
		,RESULT_SERVER_STATE_OVER, RESULT_OUTCOME_CLIENT_ERROR, RESULT_COMPUTE_DONE,
		stderr_out, FILE_DELETE_READY, VALIDATE_STATE_INVALID, exit_status, resultid);
		
	retval = boinc_db.do_query(query);
    	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
					"Failed to update result table to reflect the failure of result %d!\n", resultid);
		retval = set_status_time(MAXIMUM_DB_INT);
		if (retval)
			log_messages.printf(SCHED_MSG_LOG::CRITICAL,
						"Setting the status time for result %d failed!\n", resultid);
		return mysql_errno(boinc_db.mysql);
		
	}
	
	retval = set_status_time(MAXIMUM_DB_INT); //Making sure get_possibly_finished_job does not pick up this job.
	if (retval)
		log_messages.printf(SCHED_MSG_LOG::CRITICAL,
						"Setting the status time for result %d failed!\n", resultid);
		
	return retval;
}
 

/******************************************************************************
Name: set_job_finished

Function: Updates the 'result' table to reflect that the job has finished.
	  If the associated workunit has not finished, the 'workunit' table
	  will also be updated and the result made the canonical result of the
	  workunit. We trust the Grid. Finally the 'status_time' is set to 
	  'MAXIMUM_DB_INT' to insure get_possibly-finished_jobs does not pick
	  is up.  

Called by: bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::set_job_finished(){
	char query[MAX_QUERY_LEN];
	char xml_doc_out[LARGE_BLOB_SIZE];
	char filename[256];
	char path[272];  
	char url[256];
	char md5sum[256];
	char *pos;
	int retval;
	int max_nbytes;
	double nbytes;
	struct stat fileinfo;
		
			
	pos = strstr(xml_doc_in, "<file_info>"); //Find the first file in 'xml_doc_in'.
	
	if (!pos){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
			"Couldn't parse output file information from the BOINC database for result %d!\n", resultid);
		set_job_failed();
		return 1;
	}
	
	xml_doc_out[0] = 0; //Initialise xml_doc_out.
		
	while (pos){ //Find all the files in 'xml_doc_in'.
		filename[0] = 0;
		url[0] = 0;
		max_nbytes = 0;
		parse_str(pos, "<file_name>", filename, 256); //Parse the file-name.
		parse_str(pos, "<url>", url, 256); //Parse the url - purely cosmetic.
		parse_int(pos, "<max_nbytes>", max_nbytes); //Parse the maximum number of bytes the files as allowed to take up.
		
		if(filename[0]==0 || url[0]==0 || max_nbytes == 0){ //If we failed to parse one of the three.
			log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
				"Couldn't parse output file information from the BOINC database for result %d!\n", resultid);
			set_job_failed();
			return 1;
		}
		
		sprintf(path, "../upload/%s", filename); //Getting the relative path to the file.
		
		retval = stat(path, &fileinfo); //Getting info about the file.
		if (retval){ //The file does not exist.
			log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
					"Output file '%s' from result %d is missing - suggests a failure, check stderr_out!\n", path, resultid);
			set_job_failed();
			return 1;		
		}
		retval = md5_file(path, md5sum, nbytes); //Calculating the md5-sum of the file.
		if(retval){
			log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
					"Failed to calculate MD5-sum for '%s' from result %d!\n", path, resultid);
			md5sum[0] = 0;
		}
		
		//This sprintf from xml_doc_out to xml_doc_out is okay, because we start by printing xml_doc_out!		
		sprintf(xml_doc_out, 
			"%s<file_info>\n"
			"\t<name>%s</name>\n"
			"\t<nbytes>%d</nbytes>\n"
			"\t<max_nbytes>%d</max_nbytes>\n"
			"\t<md5_cksum>%s</md5_cksum>\n"
			"\t<url>%s</url>\n"
			"</file_info>\n"
			, xml_doc_out, filename, (int)fileinfo.st_size, max_nbytes, md5sum, url);
	
		pos = strstr(pos+11, "<file_info>"); //Get the next file.
	}
	
	
	//All the output files were there, so update the result table.
	
	sprintf(query,
    		"UPDATE result SET server_state = %d, outcome = %d, client_state = %d, received_time = %d, "
		"cpu_time = 0, xml_doc_out = '%s', validate_state = %d WHERE id = %d",
		RESULT_SERVER_STATE_OVER, RESULT_OUTCOME_SUCCESS, RESULT_FILES_UPLOADED, int(time(0)),
		xml_doc_out, VALIDATE_STATE_VALID, resultid);
	
	retval = boinc_db.do_query(query);
    	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
					"Failed to update result table to reflect the finishing of result %d!\n", resultid);
		delete_files(); //A bit of garbage collection.
		return mysql_errno(boinc_db.mysql);
	}
	
	//Make sure no more results for this workunit get sent out and that the output files get deleted when the wu has been assimilated.
	
	sprintf(query,
    		"UPDATE result SET server_state = %d, outcome = %d WHERE workunitid = %d AND server_state = %d",
		RESULT_SERVER_STATE_OVER, RESULT_OUTCOME_DIDNT_NEED, wu.id, RESULT_SERVER_STATE_UNSENT);
		
	retval = boinc_db.do_query(query);
    	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, 
					"Failed to update result table to reflect the finished of workunit %d!\n", wu.id);
		delete_files(); //A bit of garbage collection
		return mysql_errno(boinc_db.mysql);
	}
	
	//Then the workunit table, but only update it if the job has not already finished.
		
	sprintf(query,
    		"UPDATE workunit SET canonical_resultid = %d, "
		"transition_time  = %d, assimilate_state = %d "
		" WHERE id = %d AND canonical_resultid = 0",
		resultid, int(time(0)), ASSIMILATE_READY, wu.id);
	
	retval = boinc_db.do_query(query);
    	if (retval){
		delete_files(); //A bit of garbage collection.
		return mysql_errno(boinc_db.mysql);
	}
	
	delete_files(); //A bit of garbage collection.
	
	return set_status_time(MAXIMUM_DB_INT);
}


/******************************************************************************
Name: set_status_time

Function: Sets the status_time of a job by setting the GRID_DB_WORK_ITEM's 
	  variabel and by updating the 'grid_job' table.  

Called by: set_job_failed, set_job_finished and bridge_loop in bridge.C.
******************************************************************************/
int GRID_DB_WORK_ITEM::set_status_time(int new_status_time){
	char query[MAX_QUERY_LEN];
	int retval;
	
	sprintf(query,
    		"UPDATE grid_job SET status_time = %d WHERE grid_jobid = '%s'",
		new_status_time, grid_jobid);
	
	retval = boinc_db.do_query(query);
    	if (retval) return mysql_errno(boinc_db.mysql);
	
	status_time = new_status_time;
	return 0;
}


/******************************************************************************
Name: lookup (GRID_APP_VERSION)

Function: Looks up the latest version-number of an application. It also stores
	  the xml_doc and the name of the application. Exits the program if
	  it fails.

Called by: bridge_loop in bridge.C.
******************************************************************************/
void GRID_APP_VERSION::lookup(int appid){
	char query[MAX_QUERY_LEN];
    	int retval;
	MYSQL_ROW row;
	CURSOR cursor;
	
	sprintf(query, "SELECT app_version.version_num, app_version.xml_doc, app.name FROM app_version, platform, app WHERE "
		"app_version.appid=%d AND platform.name LIKE '%%linux%%' AND app.id = app_version.appid "
		"AND platform.id = app_version.platformid ORDER BY app_version.version_num DESC LIMIT 1",
		appid);
		
	retval = boinc_db.do_query(query);
	if (retval){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Couldn't lookup latest app_version no. of app %d.\n", appid);
		exit(EXIT_FAILURE);
		}
	cursor.rp = mysql_store_result(boinc_db.mysql);
	if (!cursor.rp){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Couldn't lookup latest app_version no. of app %d.\n", appid);
		exit(EXIT_FAILURE);
	}
	row = mysql_fetch_row(cursor.rp);
	if (!row) {
        	mysql_free_result(cursor.rp);
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Couldn't lookup latest app_version no. of app %d.\n", appid);
		exit(EXIT_FAILURE);
	}
	version=safe_atoi(row[0]); //Parse the app-version.
	
	safe_strncpy(xml_doc, row[1], sizeof(xml_doc)); //Parse the xml_doc.
	
	safe_strncpy(app_name, row[2], sizeof(app_name)); //Parse the name.
	 
	mysql_free_result(cursor.rp);
	
	return;
}


/******************************************************************************
Name: lookup (GRID_USER)

Function: Looks up the userid, hostid and teamid for the registered Grid user.
	  If the user is not found, it will be created.

Called by: bridge_loop in bridge.C.
******************************************************************************/
void GRID_USER::lookup(){
	char query[MAX_QUERY_LEN];
    	int retval;
	MYSQL_ROW row;
	CURSOR cursor;
	
	hostid=0;
	userid=0;
	teamid=0;
	
	safe_strcpy(query, "SELECT user.id, user.teamid, host.id FROM user, host WHERE "
		"user.id = host.userid AND email_addr='grid@lcg.org' AND name="
		"'The Almighty Grid' AND authenticator ="
		"'gridgridgridgridgridgridgridgridgridgrid' AND url"
		"= 'http://lcg.web.cern.ch/LCG/' LIMIT 0,1");	
	
	retval = boinc_db.do_query(query); //Finding the user.
	if (retval)
		goto error;
		
	cursor.rp = mysql_store_result(boinc_db.mysql);
	if (!cursor.rp)
		goto error;
	
	row = mysql_fetch_row(cursor.rp);
	
	if (row) { //The user existed, so get the info.
		userid=safe_atoi(row[0]); //Parse the userid.
		teamid=safe_atoi(row[1]); //Parse the teamid.
		hostid=safe_atoi(row[2]); //Parse the hostid.
        	mysql_free_result(cursor.rp);
		return;
	}
	
	//The user did not exist so lets create it.
	mysql_free_result(cursor.rp);
	
	log_messages.printf(SCHED_MSG_LOG::DEBUG, "Creating grid-user.\n");
	
	safe_strcpy(query, "INSERT INTO user (email_addr, name, authenticator, "
		"country, postal_code, venue, url) VALUES ('grid@lcg.org', "
		"'The Almighty Grid', 'gridgridgridgridgridgridgridgridgridgrid', "
		" 'N/A', 'N/A', 'work', 'http://lcg.web.cern.ch/LCG/')");
	
	retval = boinc_db.do_query(query); //Inserting the user into the 'user' table.
	if (retval)
		goto error;
		
	safe_strcpy(query,"SELECT LAST_INSERT_ID()"); //Getting the id of the newly inserted user.
    	retval = boinc_db.do_query(query);
    	if (retval)
		goto error;
		
    	cursor.rp = mysql_store_result(boinc_db.mysql);
    	if (!cursor.rp)
		goto error;
		
	row = mysql_fetch_row(cursor.rp);
    	if (!row) {
        	mysql_free_result(cursor.rp);
		goto error;
	}
    
    	userid=atoi(row[0]); //Parse the userid.
		
	sprintf(query, "INSERT INTO host (userid, domain_name, p_vendor, "
		"p_model, os_name, os_version) VALUES (%d, 'lcg.org', 'LCG', "
		"'So many models', 'Linux', '7.3')", userid);
			
	retval = boinc_db.do_query(query); //Inserting the host into the 'host' table.
	if (retval)
		goto error;
		
	safe_strcpy(query,"SELECT LAST_INSERT_ID()"); //Getting the id of the newly inserted host.
    	retval = boinc_db.do_query(query);
    	if (retval)
		goto error;
		
    	cursor.rp = mysql_store_result(boinc_db.mysql);
    	if (!cursor.rp)
		goto error;
		
	row = mysql_fetch_row(cursor.rp);
    	if (!row) {
        	mysql_free_result(cursor.rp);
		goto error;
    	}
    
    	hostid=atoi(row[0]); //Parse the hostid.
		
	sprintf(query, "INSERT INTO team (userid, name, url, country) VALUES "
		"(%d, 'LCG', 'http://lcg.web.cern.ch/LCG/', 'N/A')", userid);
			
	retval = boinc_db.do_query(query); //Inserting the team into the 'team' table.
	if (retval)
		goto error;
		
	safe_strcpy(query,"SELECT LAST_INSERT_ID()"); //Getting the id of the newly inserted team.
    	retval = boinc_db.do_query(query);
    	if (retval)
		goto error;
		
    	cursor.rp = mysql_store_result(boinc_db.mysql);
    	if (!cursor.rp)
		goto error;
		
	row = mysql_fetch_row(cursor.rp);
    	if (!row) {
        	mysql_free_result(cursor.rp);
		goto error;
    	}
    
    	teamid=atoi(row[0]); //Parse the teamid.
		
	//Update the user to have the correct teamid.
	sprintf(query, "UPDATE user SET teamid = %d WHERE id = %d", teamid, userid);
		
	retval = boinc_db.do_query(query);
	if (retval)
		goto error;

	return;
	
	error: //Something went wrong.
	
	log_messages.printf(SCHED_MSG_LOG::DEBUG, "Couldn't get grid-user information from the BOINC database.\n");
	return;
}
