/******************************************************************************
Name: bridge_grid.C

Description: The functions in this file handle all the actual interaction
	     with the grid. This includes submitting a job, getting the 
	     status of a job and getting the output from a job.
	     
Authors: Jakob Gregor Pedersen & Christian Ulrik Sttrup.
******************************************************************************/

#include <string>
#include "sched_msgs.h"
#include "sched_config.h"
#include "util.h"

#include "edg/workload/common/utilities/Exceptions.h"
#include "edg/workload/common/requestad/JobAd.h"
#include "edg/workload/userinterface/client/Job.h"
#include "edg/workload/logging/client/JobStatus.h"
#include "edg/workload/common/jobid/JobId.h"
#include "edg/workload/logging/client/Job.h"

using namespace std;


/******************************************************************************
Name: get_job_output

Function: Gets output-files from a grid job and stores them in a
	  specified directory. It expects to have rights to write to this 
	  directory. If the directory contains files with identical names
	  as the output-files from the grid it will fail. It will also fail
	  if valid proxy credentials are not available. 

Called by: brigde_loop in bridge.C.
******************************************************************************/
int get_job_output(char* grid_jobId, char* outputdir){
try
    {
        edg::workload::common::jobid::JobId jid (grid_jobId);
        edg::workload::userinterface::Job job(jid);
        job.getOutput(outputdir);
        return 0;
    }
    catch  (edg::workload::common::utilities::Exception &exc)
    {
        	log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to get output for grid-job %s!\n", grid_jobId);
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Stack-trace from failure:%s\n", exc.printStackTrace().c_str());
    }
    return 1;
}


/******************************************************************************
Name: get_job_status

Function: Gets the status-code, exit-code and done-code for a specified
	  grid job. If valid proxy credentials are not available it will fail.

Called by: brigde_loop in bridge.C.
******************************************************************************/
int get_job_status(char* grid_jobId, int& job_status, int& done_code, int& exit_code){
    try
    {
        edg::workload::common::jobid::JobId jid (grid_jobId);
        edg::workload::userinterface::Job job(jid);
	edg::workload::logging::client::JobStatus status = job.getStatus() ;
	done_code = status.getValInt(edg::workload::logging::client::JobStatus::DONE_CODE);
	exit_code = status.getValInt(edg::workload::logging::client::JobStatus::EXIT_CODE);
        job_status = status.status;
       	return 0;
    }
    catch  (edg::workload::common::utilities::Exception &exc)
    {
        	log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to get status for grid-job %s!\n", grid_jobId);
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Stack-trace from failure:%s\n", exc.printStackTrace().c_str());
    }
    return 1;
}


/******************************************************************************
Name: submit_result

Function: Submits a job to the grid. It expects to get the NS and LB hostnames
	  from the config-object, if this is not the case it fails.
	  It also expects to find the jdl-file in the '../grid/jobs' 
	  directory, if not so it will fail. The filename is expected to be
	  'workunit-name'.jdl.

Called by: brigde_loop in bridge.C.
******************************************************************************/
int submit_result(SCHED_CONFIG& config, char* wu_name, int wu_id, char* jobId, int jobIdSize){
	char jdlfilename[256];
	string job_id;
	
	//Printf error messages if NS or LB is not set in config.
	if (config.NS_hostname[0]==0){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "The NS-hostname for the Grid has not been set!\n");
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to submit job for workunit %d to Grid.\n", wu_id);
		return 1;
	}
	
	if (config.LB_hostname[0]==0){
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "The LB-hostname for the Grid has not been set!\n");
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to submit job for workunit %d to Grid.\n", wu_id);
		return 1;
	}
	
	sprintf(jdlfilename, "../grid/jobs/%s.jdl", wu_name);

	try{
		edg::workload::common::requestad::JobAd jab;
		jab.fromFile(jdlfilename);
    		edg::workload::userinterface::Job job(jab);
		job.setLoggerLevel (6); //What does this do????
		job.submit (config.NS_hostname, config.NS_port, config.LB_hostname, config.LB_port);
		safe_strncpy(jobId,job.getJobId()->toString().c_str(), jobIdSize);
    		return 0;
    	}
    
    	catch  (edg::workload::common::utilities::Exception &exc)
    	{
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Failed to submit job for workunit %d to Grid.\n", wu_id);
		log_messages.printf(SCHED_MSG_LOG::CRITICAL, "Stack-trace from failure:%s\n", exc.printStackTrace().c_str());
    	}
    	return 1;
}
